#!/usr/bin/env pythonx
"""Implements methods for common input/output operations"""
import os
import csv
import json
import pickle
import logging

import numpy as np
import pandas as pd

COMMA_DELIMITER = ','

MODE_WRITE = 'w'
MODE_APPEND = 'a'
MODE_WRITE_BINARY = 'wb'

logger = logging.getLogger(__name__)


def write_file(filepath, data, mode=MODE_WRITE, delimiter=COMMA_DELIMITER, raise_exc=True):
    """

    Function to write a file based on extension of the file format to be written.
    :param filepath: File location to be written.
    :param data: Data to be written into the file
    :param mode: Mode to open the file with, for writing.
    :param raise_exc: Flag for raising exception if the file already exists.
    :param delimiter: delimiter for writing csv
    :return: None
    """
    os.makedirs(os.path.dirname(filepath), exist_ok=True)
    extension = os.path.splitext(filepath)[1]

    if os.path.exists(filepath) and raise_exc and mode == MODE_WRITE:
        raise Exception(f'Attempting to overwrite file: {filepath}')

    try:
        if extension == '.json':
            with open(filepath, mode) as filepath:
                json.dump(data, filepath, indent=4)
        elif extension == '.txt':
            with open(filepath, mode) as filepath:
                filepath.write(data)
        elif extension in ['.tsv', '.csv']:
            if mode == MODE_WRITE:
                pd.DataFrame(data).to_csv(filepath, sep=delimiter,
                                          index=False, quoting=csv.QUOTE_NONNUMERIC)
            elif mode == MODE_APPEND:
                header = not os.path.exists(filepath)
                pd.DataFrame(data).to_csv(
                    filepath, sep=delimiter, mode=mode,
                    header=header, index=False, quoting=csv.QUOTE_NONNUMERIC
                )
        elif extension == '.pkl':
            with open(filepath, MODE_WRITE_BINARY) as filepath:
                pickle.dump(data, filepath, protocol=4)
        elif extension == '.npy':
            np.save(filepath, data)
        elif extension == '.hdf5':
            # As this would imply data is keras model. Since we use .hdf5 only while saving keras models
            data.save(filepath)
        else:
            raise Exception(f'Invalid extension: {extension}')

    except Exception as e:
        logger.error(
            'Unable to write file {}. {}'.format(filepath, str(e))
        )
        raise
