import logging
import numpy as np

from JarvisML.common.read.read_audio import read_audio
from JarvisML.utils.logger.log_entry_exit import log_entry_exit
from pyAudioAnalysis import audioSegmentation as aS

logger = logging.getLogger(__name__)


class RemoveSilence:

    @log_entry_exit(log_time=False)
    def run(self, data, speech_time_margin=0.5, min_audio_duration=1, read_kwargs=None, **kwargs):
        """
        Remove silence from audio files.
        :param data: Audio file paths from which silence needs to be removed.
        :param stft_win_len: window length for STFT [Parameter to PyAudioAnalysis method]
        :param stft_win_stride: stride length for STFT  [Parameter to PyAudioAnalysis method]
        :param speech_time_margin: Margin to be added to left and right end point of each speech snippet
        :param min_audio_duration: Min duration an audio must be to be considered for silence removal.
        :param read_kwargs: kwargs for reading audios
        :param kwargs: kwargs for this method
        :return:
        """

        read_kwargs = read_kwargs or {}
        raw_audios, sampling_rates = read_audio(data, **read_kwargs)
        processed_data = []
        for file_path, raw_audio, sampling_rate in zip(data, raw_audios, sampling_rates):

            length_raw_audio = len(raw_audio) if raw_audio is not None else None
            duration = length_raw_audio / sampling_rate if sampling_rate else None

            if sampling_rate and length_raw_audio:
                if duration <= 2 * speech_time_margin or duration <= min_audio_duration:
                    processed_data_unit = (raw_audio, sampling_rate)
                else:
                    # Get speech times for the raw audio
                    try:
                        speech_times = aS.silenceRemoval(
                            x=raw_audio, Fs=sampling_rate, stWin=0.025, stStep=0.01
                        )
                    except Exception as exc:
                        logger.error('PyAudioAnalysis silence removal method failed for audio: {file_path}, '
                                     'with exception: {exc}'.format(file_path=file_path, exc=exc))
                        processed_data_unit = (raw_audio, sampling_rate)
                    else:
                        speech_times = merge_ranges_with_small_diff(speech_times, speech_time_margin)

                        # Create the recording of speeched portions from time stamps accumulated above
                        speech_recording = np.array([], dtype=np.float32)
                        for ind, speech_time in enumerate(speech_times):
                            start_ind = int(speech_time[0] * sampling_rate)
                            end_ind = int(speech_time[1] * sampling_rate)
                            speech_recording = np.concatenate(
                                (speech_recording, raw_audio[start_ind: end_ind])
                            )
                        processed_data_unit = (speech_recording, sampling_rate)
            else:
                processed_data_unit = (None, None)

            processed_data.append(processed_data_unit)

        return processed_data