import warnings
import openai
from langdetect import detect
from googletrans import Translator
import numpy as np
from tqdm import tqdm
from sklearn.metrics.pairwise import cosine_similarity
from sklearn.metrics import silhouette_score
from sklearn.cluster import KMeans
from sentence_transformers import SentenceTransformer, util

CONTRACTION_DICT = {
    "don't": "do not",
    "didn't": "did not",
    "doesn't": "does not",
    "can't": "cannot",
    "isn't": "is not",
    "it's": "it is",
    "he's": "he is",
    "she's": "she is",
    "we're": "we are",
    "you're": "you are",
    "they're": "they are",
    "won't": "will not",
    "wasn't": "was not",
    "weren't": "were not",
}


# warnings.filterwarnings("ignore")
embedder = SentenceTransformer("all-MiniLM-L6-v2")
warnings.filterwarnings("ignore")


def detect_language(text):
    translator = Translator()
    try:
        # print(text)
        detected = translator.detect(text)
        return detected.lang == "en"
    except Exception as e:
        try:
            detected = detect(text)
            return detected == "en"
        except Exception as e2:
            return False


def translate_row(row):
    column_to_translate = "answer_value"
    translator = Translator()
    is_english = detect_language(row[column_to_translate])
    # progress_bar.progress( * (100 //9))
    if not is_english:
        try:
            trans = translator.translate(row[column_to_translate], dest="en").text
            return trans, True
        except Exception as e:
            return row[column_to_translate], False
    else:
        return row[column_to_translate], True


def translate_if_english(df):
    df["translated_answer_value"], df["translation_successful"] = zip(
        *df.apply(translate_row, axis=1)
    )
    return df


def get_embedding(corpus):
    corpus_embeddings = embedder.encode(corpus)
    corpus_embeddings = corpus_embeddings.cpu().numpy()  # Move to CPU before conversion
    corpus_embeddings = corpus_embeddings / np.linalg.norm(
        corpus_embeddings, axis=1, keepdims=True
    )
    return corpus_embeddings


def get_best_k(corpus_embeddings):
    silhouette_scores = []
    cluster_range = range(4, 15)
    best_silhouette_score = -1

    for n_clusters in cluster_range:
        kmeans = KMeans(n_clusters=n_clusters, random_state=0)
        kmeans.fit(corpus_embeddings)
        silhouette_avg = silhouette_score(corpus_embeddings, kmeans.labels_)
        silhouette_scores.append(silhouette_avg)

        if silhouette_avg >= best_silhouette_score:
            best_silhouette_score = silhouette_avg
            best_k = n_clusters

    return best_k


def get_most_representative_sentence(cluster_embeddings):
    similarity_matrix = cosine_similarity(cluster_embeddings)
    sentence_scores = np.sum(similarity_matrix, axis=1)
    most_representative_idx = np.argmax(sentence_scores)
    return most_representative_idx


def get_similarity(sentences1, sentences2):
    embeddings1 = embedder.encode(sentences1, convert_to_tensor=True)
    embeddings2 = embedder.encode(sentences2, convert_to_tensor=True)
    cosine_scores = util.cos_sim(
        embeddings1.cpu().numpy(), embeddings2.cpu().numpy()
    )  # Move to CPU before conversion
    return cosine_scores


def refine_clusters(clustered_sentences, representative_sentences, threshold=0.30):
    # Create cluster -1 if it doesn't exist
    if -1 not in clustered_sentences:
        clustered_sentences[-1] = []
        representative_sentences[-1] = "others"

    # Iterate over each cluster and its sentences
    for cluster_id, sentences in tqdm(
        list(clustered_sentences.items()), desc="Refining clusters"
    ):
        # if cluster_id!= -1:

        rep_sentence = representative_sentences[cluster_id]
        sentences_to_move = []

        for sentence in sentences[:]:  # Iterate over a copy of the list
            if sentence and get_similarity(sentence, rep_sentence) < threshold:
                sentences_to_move.append(sentence)

        for sentence in sentences_to_move:
            if cluster_id != -1:
                clustered_sentences[cluster_id].remove(sentence)
                clustered_sentences[-1].append(sentence)
                print(f"Moved '{sentence}' from Cluster {cluster_id} to Cluster -1")

        # Sort the remaining sentences in the cluster based on similarity to representative sentence
        clustered_sentences[cluster_id] = sorted(
            clustered_sentences[cluster_id],
            key=lambda x: get_similarity(x, rep_sentence),
            reverse=True,
        )

    # Remove empty clusters (other than -1)
    empty_clusters = [k for k, v in clustered_sentences.items() if not v and k != -1]
    for k in empty_clusters:
        del clustered_sentences[k]

    return clustered_sentences


def get_second_most_representative_sentence(cluster_embeddings, first_best_idx):
    # Initialize variables for second-best index and similarity
    second_best_idx = -1
    second_best_similarity = -1

    for i, embedding in enumerate(cluster_embeddings):
        # Skip the same index as the first best
        if i == first_best_idx:
            continue

        similarity = cosine_similarity(
            cluster_embeddings[first_best_idx].reshape(1, -1), embedding.reshape(1, -1)
        )[0][0]

        if similarity > second_best_similarity:
            second_best_similarity = similarity
            second_best_idx = i

    return second_best_idx


def get_completion(prompt, model="gpt-3.5-turbo"):
    messages = [{"role": "user", "content": prompt}]
    response = openai.ChatCompletion.create(
        model=model,
        messages=messages,
        temperature=0,  # this is the degree of randomness of the model's output
    )
    return response.choices[0].message["content"]


def clean_NA(text):
    return str(text).lower().strip().replace(".", " ").replace(",", " ")


def preprocess_data(data, contraction_dict):
    data = data.replace(contraction_dict)
    data["answer_value"] = data["answer_value"].replace("cx", "customer")
    data["answer_value"] = (
        data["answer_value"]
        .apply(clean_NA)
        .replace(["n/a", "nan", "Na", "NA", "N/A", "N/a", "na", "na.", "ni"], np.nan)
        .replace(contraction_dict)
    )
    data.dropna(inplace=True)
    return data


# Translate data
def translate_data(data):
    # translator = Translator()
    translated_df = translate_if_english(data)
    skipped_df = translated_df[translated_df["translation_successful"] == False]
    translated_df = translated_df[translated_df["translation_successful"] == True]
    return translated_df, skipped_df


# Create embeddings
def create_embeddings(data, column_name):
    embedder = SentenceTransformer("all-MiniLM-L6-v2")
    corpus = data[column_name].unique()
    corpus_embeddings = embedder.encode(corpus, convert_to_tensor=True)
    return corpus, corpus_embeddings


# Perform clustering
def perform_clustering(corpus_embeddings):
    best_k = get_best_k(corpus_embeddings)  # Assuming get_best_k is defined
    clustering_model = KMeans(n_clusters=best_k)
    clustering_model.fit(corpus_embeddings)
    return clustering_model.labels_


# Select representative sentences
def select_representative_sentences(cluster_assignment, corpus, corpus_embeddings):
    clustered_sentences = {}
    for sentence_id, cluster_id in enumerate(cluster_assignment):
        if cluster_id not in clustered_sentences:
            clustered_sentences[cluster_id] = []
        clustered_sentences[cluster_id].append(corpus[sentence_id])

    representative_sentences = {}
    for cluster_id, sentences in clustered_sentences.items():
        cluster_embeddings = [
            corpus_embeddings[i]
            for i, sentence in enumerate(corpus)
            if sentence in sentences
        ]
        idx = get_most_representative_sentence(
            cluster_embeddings
        )  # Assuming this function is defined
        representative_sentences[cluster_id] = sentences[idx]
    return representative_sentences, clustered_sentences


# Name clusters
def name_clusters(refined_clusters):
    cluster_names = {}
    for cluster_id, clusters in refined_clusters.items():
        if cluster_id != -1:
            max_elements = min(5, len(clusters))
            clusters_subset = clusters[:max_elements]
            prompt = f"""
            please give only a single meaningful name in 3-5 words to below cluster
            cluster: ```{clusters_subset}```
            name :
            """
            response = (
                get_completion(prompt).replace('"', "").strip()
            )  # Assuming get_completion is defined
            cluster_names[response] = clusters
        else:
            cluster_names["others"] = clusters

    return cluster_names


# Reassign based on similarity
def reassign_clusters(cluster_names):
    new_cluster_names = cluster_names.copy()
    for sentence in tqdm(new_cluster_names["others"]):
        best_similarity = -1
        best_rep_index = None

        for idx, cluster_name in enumerate(new_cluster_names.keys()):
            current_similarity = get_similarity(
                sentence, cluster_name
            )  # Assuming get_similarity is defined
            if current_similarity > best_similarity and current_similarity > 0.3:
                best_similarity = current_similarity
                best_rep_index = cluster_name

        if best_rep_index is not None:
            new_cluster_names[best_rep_index].append(sentence)
            new_cluster_names["others"].remove(sentence)
    return new_cluster_names


# Update DataFrame and Export
def update_dataframe_and_export(data, new_cluster_names):
    sentence_to_representative = {}
    for cluster_id, sentences in new_cluster_names.items():
        representative = cluster_id
        for sentence in sentences:
            sentence_to_representative[sentence] = representative

    # sentence_to_representative

    df = data.copy()
    df = df.rename(columns={"vpm_id": "voice_player_mission_id"})
    df["translated_answer_value_cluster"] = df["translated_answer_value"].map(
        sentence_to_representative
    )  # Assuming sentence_to_representative is defined
    df = df.sort_values(by="translated_answer_value_cluster")
    df[
        ["voice_player_mission_id", "answer_value", "translated_answer_value_cluster"]
    ].fillna("others").to_csv(f"test.csv", index=False)
    return df


def has_required_columns(df, columns):
    return all(column in df for column in columns)


# Main function below
def bucket(df,api_key):
    total_steps = 9  # Adjust this based on the number of steps in your process

    openai.api_key = api_key
    # Process Step 1: preprocess_data
    df = preprocess_data(df, CONTRACTION_DICT)
    print("Process Step 1: preprocess_data")

    # Process Step 2: translate_data
    translated_df, skipped_df = translate_data(df)
    print("Process Step 2: translate_data")

    # Process Step 3: create_embeddings
    corpus, corpus_embeddings = create_embeddings(
        translated_df, "translated_answer_value"
    )
    corpus_embeddings = corpus_embeddings.cpu().numpy()  # Move to CPU before conversion
    print("Process Step 3: create_embeddings")

    # Process Step 4: perform_clustering
    cluster_assignment = perform_clustering(corpus_embeddings)
    print("Process Step 4: perform_clustering")

    # Process Step 5: select_representative_sentences
    representative_sentences, clustered_sentences = select_representative_sentences(
        cluster_assignment, corpus, corpus_embeddings
    )
    print("Process Step 5: select_representative_sentences")

    # Process Step 6: refine_clusters
    refined_clusters = refine_clusters(clustered_sentences, representative_sentences)
    print("Process Step 6: refine_clusters")

    # Process Step 7: name_clusters
    cluster_names = name_clusters(refined_clusters)
    print("Process Step 7: name_clusters")

    # Process Step 8: reassign_clusters
    new_cluster_names = reassign_clusters(cluster_names)
    print("Process Step 8: reassign_clusters")

    # Process Step 9: update_dataframe_and_export
    processed_df = update_dataframe_and_export(df, new_cluster_names)
    print("Process Step 9: update_dataframe_and_export")

    return processed_df, skipped_df
