import os
import time
import logging
import requests
import pandas as pd

from urllib.parse import urlsplit
from io import BytesIO, DEFAULT_BUFFER_SIZE, BufferedWriter
from concurrent.futures import ThreadPoolExecutor, as_completed

from JarvisML.common.write.write_file import write_file
logger = logging.getLogger(__name__)
USER_AGENT = {'User-agent': 'Mozilla/5.0'}


class DownloadAudio:
    def __init__(
        self, df: pd.DataFrame, url_column: str, id_column: str, dest_folder: str, write_failed: bool = True,
        pipeline_logger=logger
    ):
        """

        :param filepath: Absolute path of the path to be used for downloading media
        :param url_column: Column name containing URLs for downloading media
        :param id_column: Column name containing IDs as unique identifiers for downloaded media
        :param dest_folder: Absolute path of the folder to be used for saving downloaded media
        :param delimiter: Delimiter to be used for reading the file
        :param write_failed: Flag for writing URLs and IDs of failed downloads to a file
        """
        self.data_frame = df[df[url_column].notnull()]
        self.url_col = url_column
        self.id_col = id_column
        self.dataset_dir = dest_folder
        self.write_failed = write_failed
        self.failed_filepath = os.path.join(self.dataset_dir, 'failed.csv')

        self.failed_downloads = []
        self.failed_df = None
        self.logger = pipeline_logger

        os.makedirs(self.dataset_dir, exist_ok=True)  # mkdir -p

    def populate_media_path(self, row):
        id = row[self.id_col]
        url = row[self.url_col]

        return self.get_media_path(_id=id, url=url)

    def get_media_path(self, url: str, _id: str):
        filename = os.path.splitext(urlsplit(url).path)[-1]
        return os.path.join(self.dataset_dir, f'{_id}{filename}')

    def download_single(self, url: str, _id: str):
        media_path = self.get_media_path(url=url, _id=_id)

        try:
            response = requests.get(url, headers=USER_AGENT)

            if response.ok:
                bio = BytesIO(response.content)
                with open(media_path, 'wb') as fp:
                    writer = BufferedWriter(fp)
                    while writer.write(bio.read(DEFAULT_BUFFER_SIZE)):
                        pass
                self.logger.info(f'Downloaded content from URL: {url}. id: {id}')
            else:
                self.logger.info(f"Couldn't download content from URL: {url}. id: {id}")
                self.failed_downloads.append((url, _id))
        except Exception as e:
            self.logger.info(f'Exception {e!r} occurred while downloading media.')
            self.failed_downloads.append((url, _id))

    def download_using_threads(self, max_workers: int = None):
        urls = list(self.data_frame[self.url_col])
        ids = list(self.data_frame[self.id_col])

        try:
            entry_time = time.time()

            with ThreadPoolExecutor(max_workers=max_workers) if max_workers else ThreadPoolExecutor() as executor:
                futures = [
                    executor.submit(self.download_single, url, _id)
                    for url, _id in zip(urls, ids)
                ]
                _ = [future.result() for future in as_completed(futures)]

            exit_time = time.time()
            self.logger.info('Time taken to download media: {} seconds'.format(round((exit_time - entry_time), 4)))

        except Exception as e:
            self.logger.info(f'Exception {e!r} occurred while downloading media.')

        self.failed_df = pd.DataFrame(
            self.failed_downloads, columns=[self.url_col, self.id_col]
        ) if self.failed_downloads else None

    def multiple_tries(self, max_workers: int = None, retry_limit: int = 2):
        for _ in range(retry_limit):
            self.logger.info(f'Download attempt: {_}')
            try:
                if self.failed_df is not None:
                    self.data_frame = self.failed_df.copy()
                    self.failed_downloads = []

                self.download_using_threads(max_workers=max_workers)
            except Exception as e:
                self.logger.info(f'Exception occurred: {e!r}')

                if self.write_failed and self.failed_df is not None:
                    self.log_failed_urls()

    def log_failed_urls(self):
        write_file(
            filepath=self.failed_filepath, data=self.failed_df, raise_exc=False
        )
