import time
import logging
from functools import wraps

EMPTY_STRING = ''

BASE_TEMPLATE = "'{method_name}' {{}}{{}}"
ARGS_TEMPLATE = ' : args: {}'
KWARGS_TEMPLATE = ' : kwargs: {}'

ENTER_PREFIX = 'Entering {}'
EXIT_PREFIX = "Exiting '{method_name}'"
TIME_LOG = ': Run-time: {time} seconds'


def log_entry_exit(logger=None, log_args=False, log_kwargs=False, log_time=True):
    """
    Decorator to log entry and exit of the decorated method
    """

    if not logger:
        logging.basicConfig(level=logging.INFO)
        logger = logging.getLogger(__name__)

    def actual_log_entry_exit(fn):
        """
        The actual method that does the work.
        Function within function so we can pass an argument to the decorator
        """

        @wraps(fn)
        def wrapper(*args, **kwargs):
            qualified_name = fn.__qualname__

            base_str = BASE_TEMPLATE.format(method_name=qualified_name)

            args_str = ARGS_TEMPLATE.format(args) if log_args else EMPTY_STRING
            kwargs_str = KWARGS_TEMPLATE.format(kwargs) if log_kwargs else EMPTY_STRING
            log_str = base_str.format(args_str, kwargs_str)

            logger.info(ENTER_PREFIX.format(log_str))
            entry_time = time.time()

            result = fn(*args, **kwargs)

            exit_time = time.time()
            exit_log_str = EXIT_PREFIX.format(method_name=qualified_name)
            if log_time:
                exit_log_str += TIME_LOG.format(time=round((exit_time - entry_time), 4))
            logger.info(exit_log_str)

            return result

        return wrapper

    return actual_log_entry_exit
