import pandas as pd


def get_utterances_from_word_timestamps(
        dict_word_timestamps, merge_threshold_sec=0.6, pretty_format=False
):
    """
    Function to form conversation utterances from ASR word timestamps.

    dict_word_timestamps: Dictionary with word timestamps from ASR for each channel. Required keys
            - start: start timestamp of the word
            - end: end timestamp of the word
            - confidence: prediction confidence
            - word: predicted word by ASR
            - Example:
                {
                    '0': [
                        {'confidence': 0.8702, 'word': 'hello', 'end': 4.3426, 'start': 4.2231},
                        {'confidence': 0.9895, 'word': 'good', 'end': 4.5418, 'start': 4.3431},
                        ....
                        ],
                    '1': [....]
                }
    merge_threshold_sec: max duration threshold to split consecutive words for utterance
    pretty_format: return pretty formatted string with consecutive conservation between channel

    """

    # Extract sentences from word timestamps

    utterances = []

    for k, v in dict_word_timestamps.items():

        # Identify if the consecutive words need to be merged
        df_words = pd.DataFrame(dict_word_timestamps[k])
        df_words['start_next'] = df_words['start'].shift(-1)
        df_words['if_merge'] = (df_words['start_next'] - df_words['end']).apply(
            lambda x: True if x < merge_threshold_sec else False
        )

        counter = 0
        group_index = [0]
        status_previous = df_words['if_merge'].iloc[0]

        for status in df_words['if_merge'][1:]:
            if status != status_previous and status_previous:
                pass

            elif status != status_previous or not status:
                counter += 1

            status_previous = status
            group_index.append(counter)

        df_words['group_index'] = group_index

        # Merge groups of words to form individual sentences
        df_words = df_words.groupby('group_index').apply(
            lambda x: [
                x['start'].iloc[0],
                x['end'].iloc[-1],
                ' '.join(x['word']),
                round(x['confidence'].mean(), 4)
            ]
        )

        df_words = pd.DataFrame(
            list(df_words),
            columns=['start', 'end', 'utterance', 'confidence']
        )
        df_words['channel'] = k

        utterances.append(df_words)

    # Merge sentences for each channel and sort by start timestamp
    utterances = pd.concat(
        utterances
    ).sort_values('start').reset_index(drop=True).to_dict(orient='records')

    # Get the pretty conversation style string
    if pretty_format:
        return utterances, get_pretty_utterances_string(utterances)
    else:
        return utterances
